/*!
 *  \file    ring.h
 *  \author  Wim Dolman (<a href="mailto:w.e.dolman@hva.nl">w.e.dolman@hva.nl</a>)
 *  \date    27-05-2021
 *  \version 1.1
 *
 *  \brief   Implementation of a ring buffer in C.
 *
 *  \details The compleet ring buffer is descripted in this header file.
 *           There is no accompanying c file.
 *           This ring buffer uses a struct with information about the
 *           ring buffer.
 *           The ring buffer can contain elements of any size. Multiple
 *           ring buffers can be used in the same project.
 *
 */

#ifndef __RING_H__
#define __RING_H__
#include <stdint.h>


/*! \brief  Struct for ring buffer
 *
 */
typedef struct {
  uint8_t *head;        //!< pointer to head of data in ring buffer
  uint8_t *tail;        //!< pointer to tail of data in ring buffer
  uint8_t *start;       //!< pointer to start of ring buffer
  uint8_t *end;         //!< pointer to end of ring buffer
  unsigned int tsize;   //!< datatype size of elements in ring buffer
  unsigned int size;    //!< total number of elements in ring buffer
  unsigned int count;   //!< number of buffered elements in ring buffer
} ring_t;

/*! \brief  Create ring struct and buffer
 *
 *  \param[in]  tsize  datatype size of elements in ring buffer
 *  \param[in]  size   total number of elements in ring buffer
 *
 *  \return if alright a pointer to the ring structure is returned else NULL
 */
static inline ring_t *ringCreate(unsigned int tsize, unsigned int size)
{
  ring_t *r = NULL;

  // allocate space for ring struct
  if ( (r = (ring_t *) malloc(sizeof(ring_t))) == NULL ) {
    return NULL;
  }

  // allocate space for ring buffer
  if ( (r->start = (uint8_t *) malloc(size * tsize)) == NULL ) {
    return NULL;
  }

  r->head   = r->start;
  r->tail   = r->start;
  r->end    = r->start + (size * tsize);
  r->tsize  = tsize;
  r->size   = size;
  r->count  = 0;

  return r;
}

/*! \brief  Free ring struct and buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return Void
 */
static inline void ringFree(ring_t *r)
{
  free(r->start);
  free(r);
}

/*! \brief  Get number of used elements in ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return Number of used elements
 */
static inline unsigned int ringGetCount(ring_t *r)
{
  return r->count;
}

/*! \brief  Get number of free elements in ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return Number of free elements
 */
static inline unsigned int ringGetFreeCount(ring_t *r)
{
  return (r->size - r->count);
}

/*! \brief  Check if ring buffer is empty
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return 1 when ring buffer is empty and 0 when ring buffer is not empty
 */
static inline int ringIsEmpty(ring_t *r)
{
  return (r->count == 0);
}

/*! \brief  Check if ring buffer is full
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return 1 when ring buffer is empty and 0 when ring buffer is not empty
 */
static inline int ringIsFull(ring_t *r)
{
  return (r->count == r->size);
}

/*! \brief  Write one element to ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *  \param[out] pData  Pointer to element to be written
 *
 *  \return returns 1 upon success else 0
 */
static inline unsigned int ringWrite(ring_t* r, uint8_t *pData)
{
  if ( ringIsFull(r) ) return 0;

  uint8_t *s = pData;
  uint8_t *d = r->head;

  for(int i=0; i<(r->tsize); i++) *d++ = *s++;

  (r->head) += r->tsize;
  if ( r->head >= r->end ) {
    r->head = r->start;
  }

  (r->count)++;

  return 1;
}

/*! \brief  Read one element from ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *  \param[in]  pData  Pointer to location where value of element must be written
 *
 *  \return Pointer to location where value of element must be written
 */
static inline uint8_t *ringRead(ring_t *r, uint8_t *pData)
{
  if ( ringIsEmpty(r) ) return NULL;

  uint8_t *s = r->tail;
  uint8_t *d = pData;

  for(int i=0; i<(r->tsize); i++) *d++ = *s++;

  (r->tail) += r->tsize;
  if ( r->tail >= r->end ) {
    r->tail = r->start;
  }

  (r->count)--;

  return pData;
}

/*! \brief  Peek next element of the ring buffer
 *          Read next element but keep this element in the ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *  \param[out] pData  Pointer to location where element must be written
 *
 *  \return Pointer to location where element must be written
 */
static inline uint8_t *ringPeek(ring_t* r, uint8_t *pData)
{
  uint8_t *s = r->tail;
  uint8_t *d = pData;

  for(int i=0; i<(r->tsize); i++) *d++ = *s++;

  return pData;
}


/*! \brief  Search one byte (element) in ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *  \param[in]  token  Token to be searched
 *
 *  \note   Only suitable for a  ringbuffer of tsize is 1 e.g. the type char
 *
 *  \return If token is found a pointer to the location is returned else NULL
 */
static inline uint8_t *ringSearchToken(ring_t* r, uint8_t token)
{
  uint8_t *ptr = r->tail;

  if ( r->tsize > 1 )  return NULL;

  for (int i=0; i<r->count; i++) {
    if (*ptr == token) return ptr;
    ptr++;
    if (ptr == r->end) {
      ptr = r->start;
    }
  }

  return NULL;
}

/*! \brief  Get a complete line from the ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *  \param[out] pLine  Pointer to a line buffer where data must be written to
 *
 *  \note   Only suitable for a ring buffer of tsize is 1 e.g. the type char
 *  \note   The line buffer needs to be large enough to contain the data
 *
 *  \return If line is found a pointer to location of the line buffer is returned else NULL
 */
static inline uint8_t *ringGetLine(ring_t* r, uint8_t *pLine)
{
  uint8_t *ptr;
  unsigned int n1, n2;

  if ( (ptr = ringSearchToken(r, '\n')) == NULL) {
    return NULL;
  }

  if (ptr >= r->tail) {
    n1 = ptr - r->tail;
    n2 = 0;
  } else {
    n1 = r->end - r->tail;
    n2 = ptr - r->start;
  }
  memcpy(pLine, r->tail, n1);
  memcpy(pLine + n1, r->start, n2);
  *(pLine + n1 + n2) = '\0';
  r->count -= (n1 + n2);

  r->tail = ptr;
  r->tail += r->tsize;
  if (r->tail >= r->end ) {
    r->tail = r->start;
  }
  r->count--;

  return pLine;
}

/*! \brief  Print information to stdout about the ring buffer
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \return Void
 */
static inline void ringPrintInfo(ring_t *r)
{
  printf("-------------------------------------\n");
  printf("Start  : %8lld   End   : %8lld \n", (long long int) r->start,  (long long int) r->end);
  printf("Head   : %8lld   Tail  : %8lld \n", (long long int) r->head,   (long long int) r->tail);
  printf("Free   :   %6d   Count :   %6d \n",  ringGetFreeCount(r), ringGetCount(r));
  printf("Full   :   %6d   Empty :   %6d \n",  ringIsFull(r),       ringIsEmpty(r));
  printf("-------------------------------------\n");
}

/*! \brief  Print the content in the ring buffer to stdout
 *
 *  \param[in]  r      Pointer to ring structure
 *
 *  \note   Only suitable for a ring buffer of tsize is 1 e.g. the type char
 *
 *  \return Void
 */
static inline void ringPrintBuffer(ring_t *r)
{
  uint8_t *ptr = r->tail;

  if ( r->tsize > 1 )  return;

  if ( ringIsEmpty(r) )  return;

  printf("\t");

  for (int i=0; i<r->count; i++) {
    if ( *ptr == '\n') printf("%02X\n\t", *ptr);
    else printf("%02X-", *ptr);
    ptr++;
    if (ptr == r->end) {
      ptr = r->start;
    }
  }
  printf("\n");
}


#endif
